def ceil(x: int, decimals: int) -> int:
    """
    Return the ceiling of x given the amount of decimals.
    This is the smallest integer >= x.

    :param x: any integer number
    :type x: int
    :param decimals: number of decimals
    :type decimals: int
    :return: the ceiling of x
    :rtype: int

    :raise Exception: raised when decimals is negative.
    """
    pass


def floor(x: int, decimals: int) -> int:
    """
    Return the floor of x given the amount of decimals.
    This is the largest integer <= x.

    :param x: any integer number
    :type x: int
    :param decimals: number of decimals
    :type decimals: int
    :return: the floor of x
    :rtype: int

    :raise Exception: raised when decimals is negative.
    """
    pass


def sqrt(x: int) -> int:
    """
    Gets the square root of a number.

    :param x: a non-negative number
    :type x: int
    :return: the square root of a number
    :rtype: int

    :raise Exception: raised when number is negative.
    """
    pass
