# for package imports in the smart contract
import boa3.builtin.interop.blockchain
import boa3.builtin.interop.contract
import boa3.builtin.interop.crypto
import boa3.builtin.interop.iterator
import boa3.builtin.interop.json
import boa3.builtin.interop.oracle
import boa3.builtin.interop.policy
import boa3.builtin.interop.role
import boa3.builtin.interop.runtime
import boa3.builtin.interop.stdlib
import boa3.builtin.interop.storage
