__all__ = ['OracleResponseCodeType',
           'OracleType',
           ]

from boa3.model.builtin.interop.oracle.oracleresponsecodetype import OracleResponseCodeType
from boa3.model.builtin.interop.oracle.oracletype import OracleType
