__all__ = ['JsonDeserializeMethod',
           'JsonSerializeMethod'
           ]

from boa3.model.builtin.interop.json.jsondeserializemethod import JsonDeserializeMethod
from boa3.model.builtin.interop.json.jsonserializemethod import JsonSerializeMethod
